// Copyright (C) 2014 by Jorge Munoz-Gama. All rights reserved. 
// Released under the terms of the Lesser GNU General Public License (L-GPL) version 2 or later.

package org.processmininguc.balsa.algorithms;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.processmining.models.graphbased.directed.petrinet.InhibitorNet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetFactory;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmininguc.balsa.models.ActivityDiagram;
import org.processmininguc.balsa.models.ActivityDiagramActivity;
import org.processmininguc.balsa.models.ActivityDiagramArc;
import org.processmininguc.balsa.models.ActivityDiagramElement;
import org.processmininguc.balsa.models.ActivityDiagramFinalNode;
import org.processmininguc.balsa.models.ActivityDiagramInitialNode;
import org.processmininguc.balsa.models.Artifact;
import org.processmininguc.balsa.models.Balsa;
import org.processmininguc.balsa.models.StateMachineDiagram;
import org.processmininguc.balsa.models.StateMachineDiagramCondition;
import org.processmininguc.balsa.models.StateMachineDiagramEffect;
import org.processmininguc.balsa.models.StateMachineDiagramEvent;
import org.processmininguc.balsa.models.StateMachineDiagramState;
import org.processmininguc.balsa.models.StateMachineDiagramTransition;
import org.processmininguc.balsa.models.impl.smadu.StateMachineDiagramArtCondImpl;
import org.processmininguc.balsa.models.structures.ArtifactMultRel;
import org.processmininguc.balsa.models.structures.OpType;
import org.processmininguc.balsa.models.structures.SimpleADChange;
import org.processmininguc.balsa.parameters.ConvertBalsaArtifactToPetriNetConfiguration;
import org.processmininguc.utils.PetrinetPrinter;

import com.google.common.collect.HashBiMap;

import edu.upc.essi.petrinets.uuid.ArcId;
import edu.upc.essi.petrinets.uuid.ArcType;
//import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import edu.upc.essi.petrinets.uuid.Petrinet;
import edu.upc.essi.petrinets.uuid.PlaceId;
import edu.upc.essi.petrinets.uuid.TransitionId;
import generic.Stereotype;
import generic.stereotype.EndStereotype;
import stateMachine.State;
import stateMachine.Transition.EventDepCondType;

public class BalsaArtifactToPetriNetConverter {
	
	//Initial Parameters
	private Balsa balsa;
	private ConvertBalsaArtifactToPetriNetConfiguration configuration;
	private Artifact artifactToConvert;
	
	//Map for keeping track of stereotypes
	//private Map<ArcId,EventDepCondType> m_stereotypeSMDsMap;
	//private Map<ArcId,Stereotype> m_stereotypeADsMap;
	
	//Resulting intermediate models
	private Map<String,Petrinet> m_smdList;
	private Map<String,Petrinet> m_adList;
	//private Map<String,Petrinet> m_adContextDepList;
	
	//ProM Petrinet
	Collection<InhibitorNet> m_promPNList;
	
	//Intermediate: Step 1 - Translate SMD into Petri-net
	//TODO All the elements needed for do step 1 and result of it

	
	//Intermediate: Step 2
	//TODO All the elements needed for do step 2 and result of it

	//Intermediate: Step 2 (Final)
	//TODO All the elements resulting at the end of step 2
	//List<String> stp2_places;
	//List<String> stp2_transitions;
	//List<Pair<String,String>> stp2_arcs;
	//List<Pair<String, String>> stp2_inhibitor;
	//List<String> stp2_marking;
	
	
	//Final Results
	private InhibitorNet convertedNet;
	private Marking convertedInitialMarking;
	
	//Some variables
	private final String k_availablePlace = "Available_";
	private final String k_takenPlace = "Taken_";
	
	
	public BalsaArtifactToPetriNetConverter(Balsa balsa, ConvertBalsaArtifactToPetriNetConfiguration configuration) {
		this.balsa = balsa;
		this.configuration = configuration;
		this.artifactToConvert = configuration.getArtifactToConvert();
		convertedInitialMarking = new Marking();
		m_smdList = new HashMap<>();
		m_adList = new HashMap<>();
		//m_adContextDepList = new HashMap<>();
		m_promPNList = new LinkedList<>();
		convert();
	}
	
	private void convert(){
		step1(); //Translate SMDs
		step2(); //Translate ADs
		step3(); //Incorporate ADs into SMDs
		step4(); //Translate into the Petri net format required by ProM
		for(InhibitorNet promPN: m_promPNList){
			PetrinetPrinter.print(promPN);
		}
	}
	

	/**
	 * Transforms each state machine diagram into a Petri net
	 * @author Montse
	 */
	private void step1(Artifact art){
		//StateMachineDiagram smd = artifactToConvert.getStateMachineDiagram();
		//List<StateMachineDiagram> smdList = balsa.getSMDs();
		StateMachineDiagram smd = art.getStateMachineDiagram();
		//for(StateMachineDiagram smd: smdList){
			String smdName = smd.getName();
			HashMap<String,PlaceId> placesHashMap = new HashMap<>();
			StateMachineDiagramState iniState = smd.getInitialState();
			Petrinet smdPetriNet = new Petrinet(smdName);
			
			//If it is NOT context-independent, we add the available (with the corresponding number of tokens) and the taken places
			if(!art.isContextIndependent()){
				List<ArtifactMultRel> artMultRelList = art.getContext();
				for(ArtifactMultRel artMultRel: artMultRelList){
					PlaceId availablePlaceId = smdPetriNet.addPlace("Available_" + artMultRel.getArtifactName());
					PlaceId takenPlaceId = smdPetriNet.addPlace(k_takenPlace + artMultRel.getArtifactName());
					smdPetriNet.setPlaceTokens(availablePlaceId, artMultRel.getMaxCard());
				}
			}	
			
			//getAllStatesList returns all the states in the SMD + the Initial state. It does not return the final state nor any pseudonodes.
			List<StateMachineDiagramState> stateList = smd.getAllStatesList();
			for(StateMachineDiagramState sourceState: stateList){
				//For each state, we create the corresponding place in the Petri net.
				String stateName = BalsaArtifactToPetriNetConverter.getStateName(sourceState);
				if(stateName == null){
					//TODO: ERROR!
				}
				else{
					PlaceId place;
					if(!placesHashMap.containsKey(stateName)){
						place = smdPetriNet.addPlace(stateName);
								//smdPetriNet.addPlace(stateName);
						placesHashMap.put(stateName,place);
					}
					else{
						place = placesHashMap.get(stateName);
					}
					//We also deal with the outgoing transitions for that particular state/node
					List<StateMachineDiagramTransition> transList = sourceState.getOutgoingTransition(); 
					Collections.sort(transList);
					int i=0;
					int j=1;
					String pnTransName;
					while (i<transList.size()){
						StateMachineDiagramEvent iEvent = transList.get(i).getEvent();
						if (iEvent == null){
							StateMachineDiagramEffect smdEffect = transList.get(i).getSmdEffect();
							pnTransName = smdEffect.getName();
						}
						else {
							pnTransName = transList.get(i).getEvent().getName();
						}
						TransitionId pnTrans = smdPetriNet.addTransition(pnTransName);
						smdPetriNet.addArc(place, pnTrans);
						//We obtain the target state of the transition and transform it into a place
						StateMachineDiagramState targetState = transList.get(i).getTargetNode();
						PlaceId targetPlace;
						String targetStateName = BalsaArtifactToPetriNetConverter.getStateName(targetState);
						if (!placesHashMap.containsKey(targetStateName)){
							targetPlace = smdPetriNet.addPlace(targetStateName);
							placesHashMap.put(targetStateName,targetPlace);
						}
						else {
							targetPlace = placesHashMap.get(targetStateName);
						}
						//We obtain the guard condition
						StateMachineDiagramCondition smdCondI = transList.get(i).getGuardCondition();
						ArcId arcWithStereotypeI;
						if(smdCondI != null && smdCondI.referencesArtifactNum()){
							StateMachineDiagramArtCondImpl condition = (StateMachineDiagramArtCondImpl) smdCondI;
							arcWithStereotypeI = createArcWithCondition(condition,pnTrans,targetPlace,smdPetriNet);
						}
						else{
							arcWithStereotypeI = smdPetriNet.addArc(pnTrans, targetPlace);
						}
						StateMachineDiagramEvent jEvent = null;
						StateMachineDiagramCondition smdCondJ = null;
						if(j<transList.size()){
							jEvent = transList.get(j).getEvent();
							smdCondJ = transList.get(j).getGuardCondition();
						}
						EventDepCondType tag = transList.get(i).getTag();
						if (tag!=null){
							//m_stereotypeSMDsMap.put(arcWithStereotypeI,transList.get(i).getTag());
							boolean result = smdPetriNet.addEventDepCondToArc(arcWithStereotypeI, transList.get(i).getTag());
							if(!result){
								System.out.println("Error adding the event-dependent condition");
							}
						}
						//No esta mirant la condicio de la segona transicio - si els events son iguals amb el mateix source state la condicio tb hauria de ser igual pq es null
//						if (jEvent != null  && iEvent != null && smdCondI == null && transList.get(i).getEvent().getName().equals(transList.get(j).getEvent().getName())){
//							//We add the tag to the arc
//							EventDepCondType tag = transList.get(i).getTag();
//							if (tag!=null){
//								//m_stereotypeSMDsMap.put(arcWithStereotypeI,transList.get(i).getTag());
//								boolean result = smdPetriNet.addEventDepCondToArc(arcWithStereotypeI, transList.get(i).getTag());
//								if(!result){
//									System.out.println("Error adding the event-dependent condition");
//								}
//							}
//							//TODO: No tinc clara la relacio entre aquests dos ifs (excloents /no excloents) 
//
//							else if(tag == null && smdCondI == null) {
//								//We should never reach this point
//								System.err.println("Error in the definition of conditions and tags in SMD");
//							}
//						}
						Boolean test = null;
						if(smdCondJ != null && smdCondI != null){
							test = smdCondJ.equals(smdCondI);
						}
						boolean sameEventNoCondBool = (smdCondI == null) && (smdCondJ==null) && (jEvent != null)  && (iEvent != null) && (transList.get(i).getEvent().getName().equals(transList.get(j).getEvent().getName()));
						boolean sameEventSameCondBool = (smdCondI != null) && (smdCondI.referencesArtifactNum()) && (smdCondJ != null) && (smdCondJ.equals(smdCondI)) && (jEvent != null)  && (iEvent != null) && (transList.get(i).getEvent().getName().equals(transList.get(j).getEvent().getName()));
						while(sameEventNoCondBool || sameEventSameCondBool){
							StateMachineDiagramState targetState2 = transList.get(j).getTargetNode();
							PlaceId targetPlace2;
							String targetStateName2 = BalsaArtifactToPetriNetConverter.getStateName(targetState2);
							if (!placesHashMap.containsKey(targetStateName2)){
								targetPlace2 = smdPetriNet.addPlace(targetStateName2);
								placesHashMap.put(targetStateName2,targetPlace2);
							}
							else {
								targetPlace2 = placesHashMap.get(targetStateName2);
							}
							//As we cannot add the tag of the SMD to the arc of the petri net, we create a table where we store it.

							ArcId arcWithSter = smdPetriNet.addArc(pnTrans, targetPlace2);
							EventDepCondType tagJ = transList.get(j).getTag();
							if (tagJ!=null){
								//m_stereotypeSMDsMap.put(arcWithSter, tag); //Merder tag!!!
								boolean result = smdPetriNet.addEventDepCondToArc(arcWithSter, tagJ);
								if(!result){
									System.out.println("Error adding the event-dependent condition");
								}
							}
							else if(tagJ == null && smdCondI == null) {
								//We should never reach this point
								System.err.println("Error in the definition of conditions and tags in SMD");
							}
							j++;
							if(j < transList.size()){
								jEvent = transList.get(j).getEvent();
								smdCondJ = transList.get(j).getGuardCondition();
							}
							else {
								jEvent = null;
								smdCondJ = null;
							}
							sameEventNoCondBool = (smdCondI == null) && (smdCondJ==null) && (jEvent != null)  && (iEvent != null) && (transList.get(i).getEvent().getName().equals(transList.get(j).getEvent().getName()));
							sameEventSameCondBool = (smdCondI != null) && (smdCondI.referencesArtifactNum()) && (smdCondJ != null) && (smdCondJ.equals(smdCondI)) && (jEvent != null)  && (iEvent != null) && (transList.get(i).getEvent().getName().equals(transList.get(j).getEvent().getName()));
						}
						i = j;
						j++;
					}
				}
			}
			System.out.println("Petri net array");
			System.out.println(smdPetriNet.getPlaces().toArray().toString());
			m_smdList.put(smdName,smdPetriNet);
			
		//}
		PetrinetPrinter.printAll(m_smdList);
		PetrinetPrinter.printInitialFinalPlaces(smdPetriNet);
	}
/**
 * Creates the required petri net places and transitions to ensure that the conditions in the SMD are taken into consideration by the Petrinet.	
 * @param condition
 * @param sourceTrans
 * @param targetPlace
 * @param pn
 * 
 * @return "Last" arc - the arc that leads to the target state. It will be necessary later on. Returns null if the op does nothing.
 */
private ArcId createArcWithCondition(StateMachineDiagramArtCondImpl condition, TransitionId sourceTrans, PlaceId targetPlace, Petrinet pn) {
		// TODO Auto-generated method stub
		int relatedArtNr = condition.getNumber();
		String artifactName = condition.getTargetArtifact();
		//ArcId arcToBeRemoved = pn.getArc(sourceTrans, targetPlace);
		ArcId arc = null;
		switch(condition.getMathOp()){
			case EQUALS:
				if(relatedArtNr == 0){
//					//ORIIGNAL APPROACH AS DESCRIBED IN THE PAPER
//					//Inhibitor arc at the beginning of the transition
//					PlaceId auxPlace = pn.addPlace("AuxIPlace");
//					ArcId arcTransAuxPlace = pn.addArc(sourceTrans, auxPlace);
//					TransitionId backAuxTrans = pn.addTransition("Back");
//					ArcId arcAuxPlaceBackTrans = pn.addArc(auxPlace, backAuxTrans);
//					//System.out.println("Taken Place name: " + k_takenPlace + artifactName);
//					PlaceId takenPlace = findPlaceInPN(pn, k_takenPlace + artifactName);
//					if (takenPlace == null){
//						//TODO: Throw exception
//					}
//					ArcId inhibitorArc = pn.addInhibitorArc(takenPlace,backAuxTrans);
//					ArcId arc = pn.addArc(backAuxTrans,targetPlace);
//					///Debug
//					PetrinetPrinter.printNextTransitions(takenPlace, pn);
					
					//SIMPLIFIED APPROACH
					PlaceId takenPlace = findPlaceInPN(pn, k_takenPlace + artifactName);
					if (takenPlace == null){
						//TODO: Throw exception
					}
					arc = pn.addArc(sourceTrans, targetPlace);
					ArcId inhibitorArc = pn.addInhibitorArc(takenPlace,sourceTrans);
					///Debug
					//PetrinetPrinter.printNextTransitions(takenPlace, pn);
				}
				else if (relatedArtNr == 1){
					//ORIIGNAL APPROACH AS DESCRIBED IN THE PAPER
					//Inhibitor arc at the end of the transition
					PlaceId auxPlace = pn.addPlace("AuxIPlace");
					ArcId arcTransAuxPlace = pn.addArc(sourceTrans, auxPlace);
					TransitionId backAuxTrans = pn.addTransition("Back",true);
					ArcId arcAuxPlaceBackTrans = pn.addArc(auxPlace, backAuxTrans);
					//System.out.println("Taken Place name: " + k_takenPlace + artifactName);
					PlaceId takenPlace = findPlaceInPN(pn, k_takenPlace + artifactName);
					if (takenPlace == null){
						//TODO: Throw exception
					}
					ArcId inhibitorArc = pn.addInhibitorArc(takenPlace,backAuxTrans);
					arc = pn.addArc(backAuxTrans,targetPlace);
					//pn.removeArc(arcToBeRemoved);
					///Debug
					PetrinetPrinter.printNextTransitions(takenPlace, pn);
				}
				else{
					//TODO: Throw exception
				}
				break;
			case GREATER_THAN:
				if(relatedArtNr == 0){ //Greater than zero
					//TODO: Revisar, a l'exemple no apareix!
					//Beginning
					PlaceId auxPlace = pn.addPlace("AuxIPlace");
					ArcId arcTransAuxPlace = pn.addArc(sourceTrans, auxPlace);
					TransitionId continueAuxTrans = pn.addTransition("Continue",true);
					ArcId arcAuxPlaceBackTrans = pn.addArc(auxPlace, continueAuxTrans);
					//System.out.println(k_takenPlace + artifactName);
					PlaceId takenPlace = findPlaceInPN(pn, k_takenPlace + artifactName);
					if (takenPlace == null){
						//TODO: Throw exception
					}
					ArcId takenToCont = pn.addArc(takenPlace,continueAuxTrans);
					ArcId contToTaken = pn.addArc(continueAuxTrans,takenPlace);
					arc = pn.addArc(continueAuxTrans,targetPlace);
					//Debug
					//PetrinetPrinter.printNextTransitions(takenPlace, pn);
				}
				else if (relatedArtNr == 1){ //Greater than 1
					//End
					PlaceId auxPlace = pn.addPlace("AuxIPlace");
					ArcId arcTransAuxPlace = pn.addArc(sourceTrans, auxPlace);
					TransitionId continueAuxTrans = pn.addTransition("Continue",true);
					ArcId arcAuxPlaceBackTrans = pn.addArc(auxPlace, continueAuxTrans);
					//System.out.println(k_takenPlace + artifactName);
					PlaceId takenPlace = findPlaceInPN(pn, k_takenPlace + artifactName);
					if (takenPlace == null){
						//TODO: Throw exception
					}
					ArcId takenToCont = pn.addArc(takenPlace,continueAuxTrans);
					ArcId contToTaken = pn.addArc(continueAuxTrans,takenPlace);
					arc = pn.addArc(continueAuxTrans,targetPlace);
					//pn.removeArc(arcToBeRemoved);
					//Debug
					PetrinetPrinter.printNextTransitions(takenPlace, pn);
				}
				else{
					//TODO: Throw exception
				}
				break;
			case LESS_EQUALS:
			case LESS_THAN:	
			case GREATER_EQUALS:
			default:
				//TODO: Throw exception
		}
		return arc;
		
	}
	
	private void step1(){
		for(Artifact art: balsa.getArtifacts()){
			step1(art);
		}
	}

	/**
	 * Transforms each activity diagram into a Petri net.
	 */
	private void step2(){
		//List<Petrinet> adPnList = new LinkedList();
		List<ActivityDiagram> actDiagramList = balsa.getADs();
		for (ActivityDiagram ad: actDiagramList){
			step2(ad);
			//if(!ad.getSimpleChanges().isEmpty()){
				//step2ContextDependent(ad); //Will make a copy of the ad again.
			//}
		}
		PetrinetPrinter.printAll(m_adList);
	}
	
	
	private void step2(ActivityDiagram ad){
		//Visited node List
		//HashMap<ActivityDiagramElement,PetrinetNode> visitedNodes = new HashMap<>();
		HashBiMap<ActivityDiagramElement,PlaceId> vPlacesBiMap = HashBiMap.create(); 
		HashBiMap<ActivityDiagramElement,TransitionId> vTransBiMap = HashBiMap.create(); 
		//Translate AD
		String adName = ad.getName();
		Petrinet adPetriNet = new Petrinet(adName);
		HashMap<String,PlaceId> placesHashMap = new HashMap<>();
		//getAllStatesList returns all the states in the SMD + the Initial state. It does not return the final state nor any pseudonodes.
		ActivityDiagramInitialNode adIniNode = ad.getInitialNode();
		PlaceId iniPlace = adPetriNet.addPlace("Start");
		vPlacesBiMap.put(adIniNode,iniPlace);
		
		for(ActivityDiagramArc adEdge: adIniNode.getOutgoingArcs()){
			ActivityDiagramElement adElement = adEdge.getTargetElement();
			transformADtoPN(iniPlace, adElement, adPetriNet, vPlacesBiMap, vTransBiMap);
		}

		System.out.println("Petri net array");
		System.out.println(adPetriNet.getPlaces().toArray().toString());
		m_adList.put(adName,adPetriNet);
	}
		
	
	private void addPetrinetNodesForChange(ActivityDiagram ad, Petrinet adPn, Petrinet smdPn, TransitionId source, PlaceId target) {
		ArcId arc = adPn.getArc(source, target); //Target will not be part of the SMD - this is the arc in the AD with the relevant info
		if (arc == null){
			//TODO: Throw exception
		}
		EventDepCondType evDepCond = adPn.getEventDepCondForArc(arc);
		List<SimpleADChange> changeList = findSimpleADChange(evDepCond, ad);
		for(SimpleADChange change: changeList){
			//String changedArtifact = change.getTargetArtifact();
			OpType opType = change.getOperationType();
			Collection<PlaceId> nextPlaces = smdPn.getNextPlaces(source);
			Object[] nextPlacesArray = nextPlaces.toArray(); //Had to use an array to avoid concurrent modification exception
			if(nextPlacesArray.length >= 3){
				System.out.println("Array contents");
				for(int i=0; i<nextPlacesArray.length; i++){
					System.out.println("Position " + i + " content: " + smdPn.getPlaceLabel((PlaceId)nextPlacesArray[i])); 
				}
			}
			for(int i=0; i<nextPlacesArray.length; i++){
				PlaceId nextP = (PlaceId) nextPlacesArray[i];
				ArcId arcToBeRemoved = smdPn.getArc(source, nextP);
				if (opType == OpType.INCREMENTS){
					PlaceId auxPlace = smdPn.addPlace("auxIncPlace");
					TransitionId incCountTrans = smdPn.addTransition("incCounter",true);
					//target is a final place
					smdPn.addArc(source,auxPlace);
					smdPn.addArc(auxPlace, incCountTrans);
					smdPn.addArc(incCountTrans, nextP); //No need to check if original was an inhibitor arc, as its source was a transition and its target a place.
					//smdPn.removeArc(arc);
					smdPn.removeArc(arcToBeRemoved);
					
					String avPlaceName = k_availablePlace.concat(change.getTargetArtifact());
					PlaceId availablePlace = findPlaceInPN(smdPn, avPlaceName);
					if(availablePlace == null){
						//TODO: Throw exception
					}
					
					String takenPlaceName = k_takenPlace.concat(change.getTargetArtifact());
					PlaceId takenPlace = findPlaceInPN(smdPn, takenPlaceName);
					if(takenPlace == null){
						//TODO: Throw exception
					}
					
					smdPn.addArc(availablePlace, incCountTrans);
					smdPn.addArc(incCountTrans, takenPlace);
					
				}
				else if(opType == OpType.DECREMENTS){
					PlaceId auxPlace = smdPn.addPlace("auxDecPlace");
					TransitionId decCountTrans = smdPn.addTransition("decCounter",true);
					smdPn.addArc(source,auxPlace);
					smdPn.addArc(auxPlace, decCountTrans);
					smdPn.addArc(decCountTrans, nextP); //No need to check if original was an inhibitor arc, as its source was a transition and its target a place.
					//smdPn.removeArc(arc);
					smdPn.removeArc(arcToBeRemoved);
					String avPlaceName = k_availablePlace.concat(change.getTargetArtifact());
					PlaceId availablePlace = findPlaceInPN(smdPn, avPlaceName);
					if(availablePlace == null){
						//TODO: Throw exception
					}
					
					String takenPlaceName = k_takenPlace.concat(change.getTargetArtifact());
					PlaceId takenPlace = findPlaceInPN(smdPn, takenPlaceName);
					if(takenPlace == null){
						//TODO: Throw exception
					}
					
					smdPn.addArc(takenPlace, decCountTrans);
					smdPn.addArc(decCountTrans, availablePlace);
				}
			}
		}

	}
	
	private List<SimpleADChange> findSimpleADChange(EventDepCondType evDepCond, ActivityDiagram ad){
		List<SimpleADChange> changeList = new LinkedList<>();
		for(SimpleADChange change: ad.getSimpleChanges()){
			//String targetArtifact = change.getTargetArtifact();
			EventDepCondType evDepCondCh = change.getEventDepCondType(); //It may apply only on success or failure.
			if(evDepCondCh == null || evDepCond == null || evDepCondCh.equals(evDepCond)){
				changeList.add(change);
			}
		}
		return changeList;
	}

	private void transformADtoPN(PlaceId prevPlace, ActivityDiagramElement adEl, Petrinet adPetriNet, HashBiMap<ActivityDiagramElement,PlaceId> visitedPlaces, HashBiMap<ActivityDiagramElement,TransitionId> visitedTrans){
		ArcId arc = null;
		ArcId auxArc = null;
		if(!visitedPlaces.containsKey(adEl) && !visitedTrans.containsKey(adEl)){
			ActivityDiagramElement prevActEl =  visitedPlaces.inverse().get(prevPlace);
			addNecessaryPlacesAndTransitions(prevActEl, adEl, adPetriNet, visitedPlaces, visitedTrans);
		}
		if(visitedPlaces.containsKey(adEl)){ //Both nodes are places
			//We create a transition to connect the two
			TransitionId trans = adPetriNet.addTransition("AuxTrans_"+adPetriNet.getPlaceLabel(prevPlace),true);
			auxArc = adPetriNet.addArc(prevPlace, trans);
			arc = adPetriNet.addArc(trans,visitedPlaces.get(adEl));
			
			List<ActivityDiagramElement> nextElList = adEl.getNextElements();
			for(ActivityDiagramElement newEl: nextElList){
				//If final nodes are used properly, there is no problem continuing the search
				transformADtoPN(visitedPlaces.get(adEl), newEl, adPetriNet,visitedPlaces, visitedTrans);
			}
			
		}
		else if (visitedTrans.containsKey(adEl)){ //Prev node is a place, next node is a transition
			arc = adPetriNet.addArc(prevPlace, visitedTrans.get(adEl));
			
			List<ActivityDiagramElement> nextElList = adEl.getNextElements();
			for(ActivityDiagramElement newEl: nextElList){
				//If final nodes are used properly, there is no problem continuing the search
				transformADtoPN(visitedTrans.get(adEl), newEl, adPetriNet,visitedPlaces, visitedTrans);
			}
		}
		else{ 
			//We should never reach this point
		}
		//We deal with the stereotypes
		if(adEl.isFinalNode()){
			ActivityDiagramElement prevActEl =  visitedPlaces.inverse().get(prevPlace);
			Stereotype stereo = findStereotype(prevActEl, (ActivityDiagramFinalNode) adEl);
			if(stereo != null){
				//m_stereotypeADsMap.put(arc, stereo);
				if (stereo.isEndStereotype()){
					EndStereotype endSter = (EndStereotype) stereo;
					adPetriNet.addEventDepCondToArc(arc,endSter.getValue());
					//Debugging
					System.out.println("Label " + adPetriNet.getPetrinetLabel());
					System.out.println("Previous place label " + adPetriNet.getPlaceLabel(prevPlace));
					//End debugging
					if(visitedPlaces.containsKey(adEl)){
						PlaceId place = visitedPlaces.get(adEl);
						//Debugging
						String placeName = adPetriNet.getPlaceLabel(place);
						System.out.println("Place name: "  + placeName);
						//End debugging
					}
					else if (visitedTrans.containsKey(adEl)){
						TransitionId trans = visitedTrans.get(adEl);
						//Debugging
						String transName = adPetriNet.getTransitionLabel(trans);
						System.out.println("Transition name: " + transName);
						//End debugging
					}
				}
			}
		}
	}

	private void transformADtoPN(TransitionId prevTrans, ActivityDiagramElement adEl, Petrinet adPetriNet, HashBiMap<ActivityDiagramElement,PlaceId> visitedPlaces, HashBiMap<ActivityDiagramElement,TransitionId> visitedTrans){
		ArcId arc = null;
		if(!visitedPlaces.containsKey(adEl) && !visitedTrans.containsKey(adEl)){
			ActivityDiagramElement prevActEl =  visitedPlaces.inverse().get(prevTrans);
			addNecessaryPlacesAndTransitions(prevActEl, adEl, adPetriNet, visitedPlaces, visitedTrans);
		}
		if(visitedPlaces.containsKey(adEl)){ //Both nodes are places
			//We create a transition to connect the two
			arc = adPetriNet.addArc(prevTrans, visitedPlaces.get(adEl));
			
			//Added this for debug purposes
//			List<ActivityDiagramElement> nextElListAux = adEl.getNextElements();
			
			List<ActivityDiagramElement> nextElList = adEl.getNextElements();
			for(ActivityDiagramElement newEl: nextElList){
				//If final nodes are used properly, there is no problem continuing the search
				transformADtoPN(visitedPlaces.get(adEl), newEl, adPetriNet,visitedPlaces, visitedTrans);
			}
			
		}
		else if (visitedTrans.containsKey(adEl)){ //Prev node is a place, next node is a transition
			String placeName = "auxPlace_" + adPetriNet.getTransitionLabel(prevTrans);
			PlaceId auxPlace = adPetriNet.addPlace(placeName);
			adPetriNet.addArc(prevTrans, auxPlace);
			arc = adPetriNet.addArc(auxPlace,visitedTrans.get(adEl));
			
			List<ActivityDiagramElement> nextElList = adEl.getNextElements();
			for(ActivityDiagramElement newEl: nextElList){
				//If final nodes are used properly, there is no problem continuing the search
				transformADtoPN(visitedTrans.get(adEl), newEl, adPetriNet,visitedPlaces, visitedTrans);
			}
		}
		if(adEl.isFinalNode()){
			ActivityDiagramElement prevActEl =  visitedTrans.inverse().get(prevTrans);
			//Debugging
			System.out.println("Label " + adPetriNet.getPetrinetLabel());
			System.out.println("Previous place label " + adPetriNet.getTransitionLabel(prevTrans));
			//End debugging
			Stereotype stereo = findStereotype(prevActEl, (ActivityDiagramFinalNode) adEl);
			if(stereo != null){
				//m_stereotypeADsMap.put(arc, stereo);
				if (stereo.isEndStereotype()){
					EndStereotype endSter = (EndStereotype) stereo;
					adPetriNet.addEventDepCondToArc(arc,endSter.getValue());
				}
			}
			System.out.println("Trans final node");
		}
	}
	
	/**
	 * Adds the necessary places and transitions for the given element, considering the previous element.
	 * @param prevActEl
	 * @param adEl
	 * @param adPetriNet
	 * @param visitedPlaces
	 * @param visitedTrans
	 */
	private void addNecessaryPlacesAndTransitions(ActivityDiagramElement prevActEl, ActivityDiagramElement adEl, Petrinet adPetriNet, HashBiMap<ActivityDiagramElement,PlaceId> visitedPlaces, HashBiMap<ActivityDiagramElement,TransitionId> visitedTrans){
		if(adEl.isFinalNode()){
			PlaceId place = adPetriNet.addPlace("Final");
			visitedPlaces.put(adEl,place);
		}
		else if(adEl.isActivity()){
			ActivityDiagramActivity actEl = (ActivityDiagramActivity) adEl;
			TransitionId transEl = adPetriNet.addTransition(actEl.getName());
			visitedTrans.put(adEl,transEl);
		}
		else if(adEl.isDecisionNode()){
			PlaceId curPlace = adPetriNet.addPlace("DecNode");
			visitedPlaces.put(adEl,curPlace);
		}
		else if(adEl.isMergeNode()){
			PlaceId curPlace = adPetriNet.addPlace("MergeNode");
			visitedPlaces.put(adEl,curPlace);
		}
		else{
			//TODO: Cases with fork and join nodes
			//We know it cannot be an initial node. If it is there is a huge mistake in the diagram.
			System.err.println("Error");
		}
		
	}
	
	/**
	 * "Merges" the state machine and activity diagrams
	 */
	private void step3(){
		Collection<Petrinet> smdCol = m_smdList.values();
		Iterator<Petrinet> itSmdCol = smdCol.iterator();
		while (itSmdCol.hasNext()){
			Petrinet smdPn = itSmdCol.next();
			System.out.println("Step 3 - Petrinet " + smdPn.getPetrinetLabel());
			Collection<TransitionId> pnTransSet = new LinkedList<TransitionId>(smdPn.getTransitions());
			Iterator<TransitionId> itTrans = pnTransSet.iterator();
			//Collection<String> visitedTrans = new LinkedList<>();
			while(itTrans.hasNext()){
				TransitionId transPn = itTrans.next();
				String transName = smdPn.getTransitionLabel(transPn);
				System.out.println("Transition name: " + transName);
				Petrinet adPn = m_adList.get(transName);
				if (adPn == null){
					System.err.println("There might be an error in SMD " + transName + ", could not find corresponding AD");
				}
				else{
					//Collection<PlaceId> previousPlaces = smdPn.getPreviousPlaces(transPn);
					//We have to copy it to avoid modifying the original AD, as it may be used later on by another transition
					Petrinet newAdPn = adPn.clonePetrinetWithNewIds();
					System.out.println("Printing Petri net copy");
					PetrinetPrinter.print(newAdPn);
					System.out.println("End printing Petri net copy");
					duplicatePetriNetIntoPetriNet2(smdPn,newAdPn,transPn);
				}
			}
			//Debugging purposes
			System.out.println("Aqui");
			PetrinetPrinter.print(smdPn);
			//End debug
		}
		PetrinetPrinter.printAll(m_smdList);
	}
	
	/**
	 * Translates the resulting Petri nets representing the SMDs with the embedded ADs into the Petri nets required by ProM.
	 * We assume that the initial node is the one that will have a token.
	 */
	private void step4(){
		HashMap<PlaceId,PetrinetNode> placeHashMap = new HashMap<>();
		HashMap<TransitionId,PetrinetNode> transHashMap = new HashMap<>();
		for(Petrinet smdPN: m_smdList.values()){
			InhibitorNet promPN = PetrinetFactory.newInhibitorNet(smdPN.getPetrinetLabel());
			//We create the places in the new Petrinet
			for(PlaceId smdPlace: smdPN.getPlaces()){
				//Place promPlace = promPN.addPlace(smdPN.getPlaceLabel(smdPlace).concat(smdPlace.getUUIDToString())); //Deleted
				Place promPlace = promPN.addPlace(smdPN.getPlaceLabel(smdPlace)); //Added
				if(smdPN.getInitialPlaces().contains(smdPlace) && smdPN.getPlaceLabel(smdPlace).equals("Start")){ //Obtains all the places with no input edges.
					convertedInitialMarking.add(promPlace);
				}
				else {
					for(int i=0; i<smdPN.getPlaceTokens(smdPlace); i++){ //This should probably subsume code above
					convertedInitialMarking.add(promPlace); //We have to add it to the list as many times as tokens there are.
					}
				}
				placeHashMap.put(smdPlace, promPlace);//Added
			}
			for(TransitionId smdTrans: smdPN.getTransitions()){
				//Transition promTrans = promPN.addTransition(smdPN.getTransitionLabel(smdTrans).concat(smdTrans.getUUIDToString())); //Deleted
				Transition promTrans = promPN.addTransition(smdPN.getTransitionLabel(smdTrans)); //Added
				promTrans.setInvisible(smdPN.isTransitionInvisible(smdTrans)); //Added invisible
				//System.out.println("Trans label " + smdPN.getTransitionLabel(smdTrans).concat(smdTrans.getUUIDToString()));
				transHashMap.put(smdTrans,promTrans); //Added
			}
			//For each place, we look for the next transitions and we create the link between the two in the new Petrinet.
			for(PlaceId smdPlace: smdPN.getPlaces()){
				//Place place = findPlaceInPN(promPN,smdPN.getPlaceLabel(smdPlace).concat(smdPlace.getUUIDToString())); //Deleted
				Place place = (Place) placeHashMap.get(smdPlace); //Added
				if(place == null){
					//TODO: Throw exception
					System.out.println("Error. Null place.");
				}
				for(TransitionId smdNextTrans: smdPN.getNextTransitions(smdPlace)){
					//Transition trans = findTransitionInPN(promPN, smdPN.getTransitionLabel(smdNextTrans).concat(smdNextTrans.getUUIDToString())); //Deleted
					Transition trans = (Transition) transHashMap.get(smdNextTrans); //Added
					if(trans == null){
						//TODO: Throw exception
						System.out.println("Error. Null transition: " + smdPN.getTransitionLabel(smdNextTrans).concat(smdNextTrans.getUUIDToString()));
					}
					ArcId arc = smdPN.getArc(smdPlace,smdNextTrans);
					if(smdPN.getArcType(arc) == ArcType.Inhibitor){
						promPN.addInhibitorArc(place, trans);
					}
					else{
						promPN.addArc(place, trans);
					}
				}
				//smdPN.getNextTransitions(smdPlace)
			}
			//For each transition, we look for the next places and we create the link between the two in the new Petrinet.
			for(TransitionId smdTrans: smdPN.getTransitions()){
				//Transition trans = findTransitionInPN(promPN, smdPN.getTransitionLabel(smdTrans).concat(smdTrans.getUUIDToString())); //Deleted
				Transition trans = (Transition) transHashMap.get(smdTrans); //Added
				if(trans == null){
					//TODO: Throw exception
					System.out.println("Error. Null transition.");
				}
				for(PlaceId smdNextPlace: smdPN.getNextPlaces(smdTrans)){
					//Place place = findPlaceInPN(promPN, smdPN.getPlaceLabel(smdNextPlace).concat(smdNextPlace.getUUIDToString())); //Deleted
					Place place = (Place) placeHashMap.get(smdNextPlace); //Added
					if(place == null){
						//TODO: Throw exception
						System.out.println("Error. Null place.");
					}
					promPN.addArc(trans, place);
				}
			}
			m_promPNList.add(promPN);
		}
	}

	/**
	 * Searches for a place in our own defined Petri net.
	 * @param pn
	 * @param placeName
	 * @return
	 */
	private PlaceId findPlaceInPN(Petrinet pn, String placeName) {
		Collection<PlaceId> placeList = pn.getPlaces();
		for(PlaceId place: placeList){
			if(pn.getPlaceLabel(place).equals(placeName)){
				return place;
			}
		}
		return null;
	}
	
	
	/**
	 * Searches for a place in the ProM Petrinet.	
	 * @param promPN
	 * @param placeName
	 * @return
	 */
	private Place findPlaceInPN(InhibitorNet promPN, String placeName) {
		Collection<Place> placeList = promPN.getPlaces();
		for(Place place: placeList){
			if(place.getLabel().equals(placeName)){
				return place;
			}
		}
		return null;
	}
	
	private Transition findTransitionInPN(InhibitorNet promPN, String transName){
		System.out.println("Transition name " + transName);
		Collection<Transition> transList = promPN.getTransitions();
		for(Transition trans: transList){
			if(trans.getLabel().equals(transName)){
				return trans;
			}
		}
		return null;
	}


	enum DuplicateResult{
		SUCCESS,
		BOTH_ARE_TRANSITIONS,
		BOTH_ARE_PLACES,
		N1_NOT_IN_PN1,
		N2_NOT_IN_PN2
	}
	
	/**
	 * 
	 * @param containerPn
	 * @param containedPn
	 * @param transToBeRepl
	 */
	public void duplicatePetriNetIntoPetriNet2(Petrinet containerPn, Petrinet containedPn, TransitionId transToBeRepl){
		//We obtain the places where the "last" transition of the activity diagram will have to be connected to.
		System.out.println("Duplicate Petrinet into Petrinet");
		System.out.println("Transition to be replaced: " + containerPn.getTransitionLabel(transToBeRepl));
		Collection<PlaceId> smdNextPlaces = containerPn.getNextPlaces(transToBeRepl);
		for(PlaceId place: smdNextPlaces){
			System.out.println("Next places:" + containerPn.getPlaceLabel(place));
		}
		//We obtain the places that lead to the current transition that we wish to replace
		Collection<PlaceId> smdPrevPlaces = containerPn.getPreviousPlaces(transToBeRepl);
		
		Collection<TransitionId> transCol = new LinkedList<>(containedPn.getTransitions());
		Collection<PlaceId> placeCol = new LinkedList<>(containedPn.getPlaces());
		
		placeCol.removeAll(containedPn.getInitialPlaces());
		placeCol.removeAll(containedPn.getFinalPlaces());
		
		for(TransitionId trans: transCol){
			containerPn.addTransitionFromPetrinet(trans, containedPn); //Creates all transitions in the containerPN
		}
		for(PlaceId place: placeCol){
			containerPn.addPlaceFromPetrinet(place, containedPn); //Creates all places in the containerPN
		}
		
		Collection<PlaceId> iniPlaces = containedPn.getInitialPlaces();
		if(iniPlaces.size()!=1){
			//TODO: Error - Throw exception
		}
		Iterator<PlaceId> it = iniPlaces.iterator();
		PlaceId iniPlaceAD = it.next();
		Collection<TransitionId> adNextTrans = containedPn.getNextTransitions(iniPlaceAD);
		
		//Debugging purposes
//		for(TransitionId trans: adNextTrans){
//			System.out.println("Transition a la que connectem des del node inicial: " + containedPn.getTransitionLabel(trans));
//		}
		//End debugging
		
		//Creates "initial" arcs
		for(PlaceId pl: smdPrevPlaces){
			for(TransitionId targetTrans: adNextTrans){
				ArcId oldArc = containerPn.getArc(pl,transToBeRepl);
				if(containerPn.getArcType(oldArc) == ArcType.Inhibitor){
					containerPn.addInhibitorArc(pl, targetTrans);
				}
				else{
					containerPn.addArc(pl, targetTrans);
				}
			}
		}

		for(TransitionId trans: containedPn.getTransitions()){
			Collection<PlaceId> nextPlaces = containedPn.getNextPlaces(trans);
			//Debugging purposes
//			System.out.println("Connecting nodes - Transition " + containedPn.getTransitionLabel(trans));
//			for(PlaceId place: nextPlaces){
//				System.out.println("Place: " + containedPn.getPlaceLabel(place));
//			}
			//End debug
			for(PlaceId pId: nextPlaces){
				if(placeCol.contains(pId)){
					//It is not the final place of the activity diagram
					containerPn.addArc(trans,pId);
				}
				else{
					//We get the arc between the transition and the final place of the AD:
					ArcId lastArcAD = containedPn.getArc(trans,pId);
					if (lastArcAD == null){
						System.out.println("AD arc null");
					}
					//Final place of the activity diagram. Will have to connect to the TARGET state of the transition in the SMD
					for(PlaceId smdPlace: smdNextPlaces){
						//Arc between the transition that we wish to replace and the current nextPlace.
						ArcId arcSMD = containerPn.getArc(transToBeRepl, smdPlace);
						if(arcSMD == null){
							System.out.println("SMD arc null");
						}
//						System.out.println("Debugging purposes: ");
//						System.out.println("ArcSMD " + arcSMD.getUUIDToString() + " objectID: " + arcSMD);
//						for (Map.Entry<ArcId, EventDepCondType> entry : m_stereotypeSMDsMap.entrySet()) {
//						    System.out.println(entry.getKey().getUUIDToString()+ " objectID: " + entry.getKey()+ " : "+entry.getValue());
//						}
//						System.out.println("End debugging");
//						System.out.println("Debugging purposes: ");
//						System.out.println("ArcAD " + lastArcAD.getUUIDToString() + " objectID: " + lastArcAD);
//						for (Map.Entry<ArcId, Stereotype> entry : m_stereotypeADsMap.entrySet()) {
//						    System.out.println(entry.getKey().getUUIDToString()+ " objectID: " + entry.getKey()+ " : "+entry.getValue());
//						}
//						System.out.println("End debugging");
						if(containerPn.getEventDepCondForArc(arcSMD) != null){
							System.out.println("SMD te estereotip");
						}
						if(containedPn.getEventDepCondForArc(lastArcAD) != null){
							System.out.println("AD te estereotip");
						}
//						if(m_stereotypeSMDsMap.containsKey(arcSMD)){ 
//							System.out.println("SMD te estereotip");
//						}
//						if(m_stereotypeADsMap.containsKey(lastArcAD)){
//							System.out.println("AD te estereotip");
//						}
						if((containerPn.getEventDepCondForArc(arcSMD) != null) && (containedPn.getEventDepCondForArc(lastArcAD) != null)){
							System.out.println("Hi ha clau dels dos elements");
							//EventDepCondType smdEvDepCond = m_stereotypeSMDsMap.get(arcSMD);
							if(containerPn.getEventDepCondForArc(arcSMD).equals(containedPn.getEventDepCondForArc(lastArcAD))){
									System.out.println("son iguals");
									containerPn.addArc(trans,smdPlace);
							}
						}
//						if(m_stereotypeSMDsMap.containsKey(arcSMD) && m_stereotypeADsMap.containsKey(lastArcAD)){
//							System.out.println("Hi ha clau dels dos elements");
//							if(m_stereotypeADsMap.get(lastArcAD).isEndStereotype()){
//								EndStereotype endSter = (EndStereotype) m_stereotypeADsMap.get(lastArcAD);
//								EventDepCondType adEvDepCond = endSter.getValue();
//								EventDepCondType smdEvDepCond = m_stereotypeSMDsMap.get(arcSMD);
//								if(endSter.getValue().equals(m_stereotypeSMDsMap.get(arcSMD))){
//									System.out.println("son iguals");
//									containerPn.addArc(trans,smdPlace);
//								}
//							}
//						}
						else if((containerPn.getEventDepCondForArc(arcSMD) == null) && (containedPn.getEventDepCondForArc(lastArcAD) == null)){
							containerPn.addArc(trans,smdPlace);
						}
//						else if(!m_stereotypeSMDsMap.containsKey(arcSMD) && !m_stereotypeADsMap.containsKey(lastArcAD)) {
//							//We ignore the place of the activity diagram and we connect the transition of the ad (now copied to the smd) to the place of the SMD
//							containerPn.addArc(trans,smdPlace);
//						}
						else if((containerPn.getEventDepCondForArc(arcSMD) == null) && (containedPn.getEventDepCondForArc(lastArcAD) != null)){
							containerPn.addArc(trans,smdPlace);
						}
//						else if(!m_stereotypeSMDsMap.containsKey(arcSMD) && m_stereotypeADsMap.containsKey(lastArcAD)){
//							//AD has stereotype but SMd doesnt - this means that the outcome does not impact the evolution of the artifact
//							containerPn.addArc(trans,smdPlace);
//							//System.out.println("MISSING IF");
//						}
					}
				}
			}		
		}
		for(PlaceId place: placeCol){
			Collection<TransitionId> nextTrans = containedPn.getNextTransitions(place);
			//Debugging purposes
			System.out.println("Connecting nodes - Places " + containedPn.getPlaceLabel(place));
			for(TransitionId trans: nextTrans){
				System.out.println("Trans: " + containedPn.getTransitionLabel(trans));
			}
			//End debug
			for(TransitionId trId: nextTrans){
				containerPn.addArc(place,trId);
			}
		}
		containerPn.removeTransition(transToBeRepl);
		
		
		//This part of the code deals with adding the necessary places and transitions to increment or decrement the number of tokens, according to whatever the activity diagram does.
		Artifact art = balsa.findArtifactByName(containerPn.getPetrinetLabel());
		if(art == null){
			//TODO: Throw exception
			//Debug
			//System.out.println("Step 3 - null artifact");
		}
		if(!art.isContextIndependent()){
			ActivityDiagram correspondingAD = null;
			for(ActivityDiagram ad: balsa.getADs()){
				if(ad.getName().equals(containedPn.getPetrinetLabel())){
					correspondingAD = ad;
					break;
				}
			}
			if (correspondingAD == null){
				//TODO: Throw exception
				//Debug
				//System.out.println("Step 3 - null corresponding AD");
			}
			
			Collection<PlaceId> finalPlaceList = containedPn.getFinalPlaces();
			for(PlaceId cPlace: finalPlaceList){
				Collection<TransitionId> transList = containedPn.getPreviousTransitions(cPlace);
				for(TransitionId cTrans: transList){
					//Buscar xarxa petri
					addPetrinetNodesForChange(correspondingAD, containedPn, containerPn, cTrans, cPlace);
				}
			}
		}
		
	}
	
	
	/**
	 * Created this function to "code" the name of the places.
	 * @param state
	 * @return
	 */
	private static String getStateName(StateMachineDiagramState state){
		if(state.isInitialPseudostate()){
			return "Start";
			//TODO: Should have initial marking
		}
		else if(state.isFinalState()){
			return "End";
		}
		else if(state.isState()){
			return ((State)state).getName();
		}
		else{
			return null;
		}
	}

	public InhibitorNet getConvertedNet() {
		//return this.convertedNet;
		//InhibitorNet inhNet = new InhibitorNet();
		for(InhibitorNet promPN: m_promPNList){
			if(artifactToConvert.getName().equals(promPN.getLabel())){
				return promPN;
			}
		}
		return null;
	}

	public Marking getConvertedInitialMarking() {
		return this.convertedInitialMarking;
	}
	
	private Stereotype findStereotype(ActivityDiagramElement prevNode, ActivityDiagramFinalNode finalNode){
		List<ActivityDiagramArc> incomingArcList = finalNode.getIncomingArcs();
		boolean found = false;
		int i = 0;
		Stereotype tag = null;
		while (!found && i<incomingArcList.size()){
			ActivityDiagramArc currentArc = incomingArcList.get(i);
			if(currentArc.getSourceElement() == prevNode){
				found = true;
				tag = currentArc.getStereotype();
			}
			i++;
		}
		return tag;
	}
		

}
