// Copyright (C) 2014 by Jorge Munoz-Gama. All rights reserved. 
// Released under the terms of the Lesser GNU General Public License (L-GPL) version 2 or later.

package org.processmininguc.balsa.models;

import java.util.List;

/**
 * Represents a node in the activity diagram. This may be a task (activity), a decision node, final node, etc.
 * @author Montse
 *
 */
public interface ActivityDiagramElement {
	
	/**
	 * Returns true if the ActivityDiagramElement is an activity or task. Returns false otherwise. 
	 * @return
	 */
	public boolean isActivity();
	
	/**
	 * Returns true if the ActivityDiagramElement is a decision node. Returns false otherwise. 
	 * @return
	 */
	public boolean isDecisionNode();
	
	/**
	 * Returns true if the ActivityDiagramElement is a merge node. Returns false otherwise. 
	 * @return
	 */
	public boolean isMergeNode();
	
	/**
	 * Returns true if the ActivityDiagramElement is a final node. Returns false otherwise. 
	 * @return
	 */
	public boolean isFinalNode();
	
	/**
	 * Returns true if the ActivityDiagramElement is an initial node. Returns false otherwise. 
	 * @return
	 */
	public boolean isInitialNode();
	
	/**
	 * Returns a list of the incoming ActivityDiagramArcs.
	 * @return
	 */
	public List<ActivityDiagramArc> getIncomingArcs();
	
	/**
	 * Returns a list of the outgoing ActivityDiagramArcs.
	 * @return
	 */
	public List<ActivityDiagramArc> getOutgoingArcs();
	
	/**
	 * Returns a list of the previous ActivityDiagramElements.
	 * @return
	 */
	public List<ActivityDiagramElement> getPreviousElements();
	
	/**
	 * Returns a list of the next ActivityDiagramElements.
	 * @return
	 */
	public List<ActivityDiagramElement> getNextElements();

}
