package org.processmininguc.balsa.models.impl.smadu;

import java.util.List;

import activity.Action;
import activity.Activity;
import activity.ActivityEdge;
import activity.ActivityFinalNode;
import activity.ActivityNode;
import activity.ControlFlow;
import activity.DecisionNode;
import activity.InitialNode;
import activity.MergeNode;
import generic.Note;
import generic.Stereotype;
import parser.ActivityFactoryInterface;

public class ActivityDiagramFactory implements ActivityFactoryInterface {

	public Action createAction() {
		return new ActivityDiagramActivityImpl();
	}

	public Action createAction(String name) {
		return new ActivityDiagramActivityImpl(name);
	}

	public Action createAction(String name, Activity act) {
		return new ActivityDiagramActivityImpl(name,act);
	}

	public Activity createActivity() {
		return new ActivityDiagramImpl();
	}

	public Activity createActivity(String name) {
		return new ActivityDiagramImpl(name);
	}

	public Activity createActivity(String name, ActivityDiagramInitialNodeImpl iniNode) {
		return new ActivityDiagramImpl(name, iniNode);
	}

	public ActivityFinalNode createActivityFinalNode() {
		return new ActivityDiagramFinalNodeImpl();
	}

	public ControlFlow createControlFlow() {
		return new ActivityDiagramArcImpl();
	}

	public ControlFlow createControlFlow(ActivityNode source, ActivityNode target) {
		return new ActivityDiagramArcImpl(source,target);
	}

	public ControlFlow createControlFlow(ActivityNode source, ActivityNode target, Stereotype st) {
		return new ActivityDiagramArcImpl(source,target,st);
	}

	public ControlFlow createControlFlow(ActivityNode source, ActivityNode target, Stereotype st, String guard) {
		return new ActivityDiagramArcImpl(source,target,st,guard);
	}

	public DecisionNode createDecisionNode() {
		return new ActivityDiagramDecisionNodeImpl();
	}

	public DecisionNode createDecisionNode(ActivityEdge incoming, ActivityEdge outgoing) {
		return new ActivityDiagramDecisionNodeImpl(incoming,outgoing);
	}

	public DecisionNode createDecisionNode(ActivityEdge incoming, List<ActivityEdge> outgoing) {
		return new ActivityDiagramDecisionNodeImpl(incoming,outgoing);
	}

	public InitialNode createInitialNode() {
		return new ActivityDiagramInitialNodeImpl();
	}

	public InitialNode createInitialNode(ControlFlow outgoing) {
		return new ActivityDiagramInitialNodeImpl(outgoing);
	}

	public InitialNode createInitialNode(List<ControlFlow> outgoing) {
		return new ActivityDiagramInitialNodeImpl(outgoing);
	}

	public MergeNode createMergeNode() {
		return new ActivityDiagramMergeNodeImpl();
	}

	public MergeNode createMergeNode(ActivityEdge incoming, ActivityEdge outgoing) {
		return new ActivityDiagramMergeNodeImpl(incoming,outgoing);
	}

	public MergeNode createMergeNode(List<ActivityEdge> incoming, ActivityEdge outgoing) {
		return new ActivityDiagramMergeNodeImpl(incoming,outgoing);
	}

	/**
	 * Creates an activity diagram (i.e. activity in MetaSMADU).
	 */
	public Activity createActivity(String name, InitialNode iniNode) {
		return new ActivityDiagramImpl(name, iniNode);
	}

	public Activity createActivity(String name, InitialNode iniNode, Note note) {
		return new ActivityDiagramImpl(name, iniNode, note);
	}

	public Activity createActivity(String name, InitialNode iniNode, List<Note> noteList) {
		return new ActivityDiagramImpl(name, iniNode, noteList);
	}
	

}
