package org.processmininguc.balsa.models.impl.smadu;

import java.util.LinkedList;
import java.util.List;

import org.processmininguc.balsa.models.ActivityDiagram;
import org.processmininguc.balsa.models.Annotation;
import org.processmininguc.balsa.models.Artifact;
import org.processmininguc.balsa.models.Balsa;
import org.processmininguc.balsa.models.StateMachineDiagram;
import org.processmininguc.balsa.models.structures.RestrictionType;
import org.processmininguc.parser.AnnotationParser;
import org.processmininguc.parser.ParsedAnnotation;

public class BalsaImpl implements Balsa{

	private List<Artifact> m_artifacts;
	private List<StateMachineDiagram> m_smds;
	private List<ActivityDiagram> m_ads;
	
	protected BalsaImpl(){
		m_artifacts = new LinkedList<Artifact>();
		m_smds = new LinkedList<StateMachineDiagram>();
		m_ads = new LinkedList<ActivityDiagram>();
	}
	
	protected BalsaImpl(List<Artifact> artList, List<StateMachineDiagram> smdList, List<ActivityDiagram> adList) {
		m_artifacts = new LinkedList<Artifact>(artList);
		m_smds = new LinkedList<StateMachineDiagram>(smdList);
		m_ads = new LinkedList<ActivityDiagram>(adList);
		processAnnotations();
	}

//	private void postProcess(List<StateMachineDiagram> smdList) {
//		for (StateMachineDiagram smd: smdList){
//			List<StateMachineDiagramTransition> transList = smd.getTransitionList();
//			for (StateMachineDiagramTransition trans: transList){
//				if(trans.get)
//			}
//		}
//		
//	}

	/**
	 * Creates a BALSA model with the given state machine and activity diagrams. From these lists, it creates the artifacts
	 * and parses the annotations in the activity diagram to include their information in the model. To do so, it ensures that there 
	 * the artifacts names in the annotations correspond to artifacts in the model.
	 * @param smdList
	 * @param adList
	 */
	protected BalsaImpl(List<StateMachineDiagram> smdList, List<ActivityDiagram> adList) {
		m_artifacts = new LinkedList<Artifact>();
		m_smds = new LinkedList<StateMachineDiagram>(smdList);
		m_ads = new LinkedList<ActivityDiagram>(adList);	
		processAnnotations();
	}



	public List<Artifact> getArtifacts() {
		return m_artifacts;
	}
	
	public Artifact getArtifact(String artName){
		Artifact artRes = null;
		for(Artifact art: m_artifacts){
			if(art.getName().toLowerCase().equals(artName.toLowerCase())){
				artRes = art;
				break;
			}
		}
		return artRes;
	}

	public void addArtifact(Artifact artifact) {
		// TODO Auto-generated method stub
		m_artifacts.add(artifact);
	}
	
	public List<StateMachineDiagram> getSMDs(){
		return m_smds;
	}
	
	public void addStateMachineDiagram(StateMachineDiagram smd){
		m_smds.add(smd);
	}
	
	public List<ActivityDiagram> getADs(){
		return m_ads;
	}
	
	public void addActivityDiagram(ActivityDiagram ad){
		m_ads.add(ad);
	}
	
	/**
	 * Returns true if the BALSA model contains an artifact with the given name.
	 * @param artName
	 * @return
	 */
	public boolean containsArtifact(String artName){
		boolean result = false;
		for(Artifact art: m_artifacts){
			if(art.getName().toLowerCase().equals(artName.toLowerCase())){
				result = true;
				break;
			}
		}
		return result;
	}
	
	/**
	 * Looks for the artifact with the given name. If it cannot be found, returns null.
	 */
	public Artifact findArtifactByName(String name){
		for(Artifact art: m_artifacts){
			if(art.getName().equals(name)){
				return art;
			}
		}
		return null;
	}
	
	public void processAnnotations(){
		for (ActivityDiagram ad: m_ads){
			for (Annotation annotation: ad.getNotes()){
				AnnotationParser ap = new AnnotationParser();
				ParsedAnnotation parsedAnno = ap.parseAnnotation(annotation);
				System.out.println("Hola");
				if (parsedAnno != null && parsedAnno.getRestrictionType() == RestrictionType.OPCHANGE && containsArtifact(parsedAnno.getTargetArtifact()) && containsArtifact(parsedAnno.getContext())){
					if (ad instanceof org.processmininguc.balsa.models.impl.smadu.ActivityDiagramImpl){ //TODO: Should be improved
						ActivityDiagramImpl adImpl = (ActivityDiagramImpl) ad;
						adImpl.createChangeAddToList(parsedAnno.getContext(),parsedAnno.getTargetArtifact(), parsedAnno.getOpType(), parsedAnno.getNumber(), parsedAnno.getEventDepCondType());
					}
				}
				else if(parsedAnno != null && parsedAnno.getRestrictionType() == RestrictionType.MAXMULT && containsArtifact(parsedAnno.getTargetArtifact()) && containsArtifact(parsedAnno.getContext())){
					Artifact contextArtifact = getArtifact(parsedAnno.getContext());
					if (contextArtifact instanceof org.processmininguc.balsa.models.impl.smadu.ArtifactImpl){ //TODO: Should be improved
						ArtifactImpl contextArtifactImpl = (ArtifactImpl) contextArtifact;
						contextArtifactImpl.addArtifactMultRel(parsedAnno.getTargetArtifact(), parsedAnno.getNumber());
					}
				}
			}
		}
	}
}
