package org.processmininguc.balsa.plugins;

import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.LinkedList;
import java.util.List;

import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;

import org.processmining.contexts.uitopia.annotations.UIImportPlugin;
import org.processmining.framework.abstractplugins.AbstractImportPlugin;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmininguc.balsa.models.ActivityDiagram;
import org.processmininguc.balsa.models.Balsa;
import org.processmininguc.balsa.models.StateMachineDiagram;
import org.processmininguc.balsa.models.impl.smadu.ActivityDiagramFactory;
import org.processmininguc.balsa.models.impl.smadu.ActivityDiagramImpl;
import org.processmininguc.balsa.models.impl.smadu.AnnotationFactory;
import org.processmininguc.balsa.models.impl.smadu.BalsaFactorySMADU;
import org.processmininguc.balsa.models.impl.smadu.StateMachineDiagramFactory;
import org.processmininguc.balsa.models.impl.smadu.StateMachineDiagramImpl;
import org.processmininguc.utils.Utils;

import activity.Activity;
import parser.ParserError;
import parser.ParserErrorType;
import parser.ParserFacade;
import stateMachine.StateMachine;


@Plugin(name = "Import Balsa from file", parameterLabels = { "Filename" }, 
returnLabels = { "Balsa" }, returnTypes = { Balsa.class })

@UIImportPlugin(description = "Balsa files", extensions = { "balsa" })


public class ImportBalsa extends AbstractImportPlugin {
	
	protected FileFilter getFileFilter() {
		return new FileNameExtensionFilter("Balsa files", "balsa");
	}

	protected Object importFromStream(PluginContext context, InputStream input, String filename, long fileSizeInBytes)
			throws Exception {
		
		Balsa balsa = null;
		
		BalsaFactorySMADU balsaFactory = new BalsaFactorySMADU();
		StateMachineDiagramFactory smdFactory = new StateMachineDiagramFactory();
		ActivityDiagramFactory adFactory = new ActivityDiagramFactory();
		AnnotationFactory annFactory = new AnnotationFactory();
		
		List<StateMachine> smdList = new LinkedList<StateMachine>();
		List<Activity> adList = new LinkedList<Activity>();
		
		
		File tempFile = File.createTempFile("tempFile",".tmp");
        OutputStream outStream = new FileOutputStream(tempFile);
        byte[] buffer = new byte[input.available()];
        input.read(buffer);
        outStream.write(buffer);
        outStream.flush();
        outStream.close();
        input.close();
        
		ParserError peSMD = ParserFacade.parseStateMachines(tempFile, smdList, smdFactory);
		if (!peSMD.getErrorType().equals(ParserErrorType.SUCCESS)){
			//TODO: Improve this
			System.err.println("Error SMDs");
		}
		ParserError peAD = ParserFacade.parseActivityDiagrams(tempFile, adList, adFactory, annFactory);
		if(!peAD.getErrorType().equals(ParserErrorType.SUCCESS)){
			//TODO: Improve this
			System.err.println("Error ADs");
		}
		List<ActivityDiagramImpl> adListSMADU = Utils.convert(adList);
		List<StateMachineDiagramImpl> smdListSMADU = Utils.convert(smdList);
		
		List<StateMachineDiagram> smdListI = Utils.convert(smdListSMADU);
		List<ActivityDiagram> adListI = Utils.convert(adListSMADU);
		
		balsa = balsaFactory.createBalsa(smdListI, adListI);
 			
		return balsa;
	}

}
