package org.processmininguc.parser;

import org.processmininguc.balsa.models.Annotation;
import org.processmininguc.balsa.models.structures.OpType;
import org.processmininguc.balsa.models.structures.RestrictionType;

import stateMachine.Transition.EventDepCondType;

public class AnnotationParser {
	
	private static final String contextContentSep = ":";
	private static final String contextKeyword = "context";
	private static final String invariantKeyword = "inv";
	private static final String maxMultKeyword = "maxMult";
	private static final String selfRefKeyword = "self";
	private static final String elSeparator = ".";
	private static final String sizeKeyword = "size";
	private static final String evDepCondImp = "->";
	
	
	public ParsedAnnotation parseAnnotation(Annotation note){
		//We obtain the content of the note. Should be in OCL.
		ProcessedAnnotationContext pa = getNoteContext(note);
		ParsedAnnotation parsedAnnotation = null; 
		if (pa.isSuccessful()){
			//We continue parsing
			ProcessedAnnotationDetails paDetails = getNoteDetails(note);
			if (paDetails.isSuccessful()){
				//ParsedAnnotation parsedAnn = new ParsedAnnotation(context, targetArtifact, opType, evDepCond, number, resType)
				parsedAnnotation = new ParsedAnnotation(pa.getContext(), paDetails.getTargetArtifact(), paDetails.getOpType(),paDetails.getEventDepCond(),paDetails.getValue(), pa.getRestrictionType());
			}
			else {
				//Return error
			}
		}
		else {
			//We should return an error message.
		}
		return parsedAnnotation;
	}
	

	/**
	 * 
	 * @param note
	 * @return
	 */
	private ProcessedAnnotationContext getNoteContext(Annotation note){
		String noteText = note.getText();
		String [] splitContextContent = noteText.split(contextContentSep);
		//First part of vector should contain the context
		if (splitContextContent.length != 2){
			return new ProcessedAnnotationContext("",false,"Badly defined expression. Revise the use of " + contextContentSep,null);
			//return new ProcessedAnnotation("",false,"Missing separator " + contextContentSep + "  keyword",null);
		}
		else {
			String [] contextSplit = splitContextContent[0].split("(?i)" + contextKeyword); //To make it case insensitive
			if (contextSplit.length != 2){
				return new ProcessedAnnotationContext("",false,"Missing " + contextKeyword + "  keyword",null);
			}
			else if(contextSplit[1].toLowerCase().contains(invariantKeyword.toLowerCase())){
				String [] invSplit = contextSplit[1].split("(?i)" + invariantKeyword);
				if (invSplit.length != 1){
					return new ProcessedAnnotationContext("",false,"Badly defined " + invariantKeyword + "  keyword",null);
				}
				else{
					String contextName = invSplit[0].trim();
					return new ProcessedAnnotationContext(contextName,true,"Found change",RestrictionType.OPCHANGE);
				}
			}
			else if(contextSplit[1].toLowerCase().contains(maxMultKeyword.toLowerCase())){
				String [] invSplit = contextSplit[1].split("(?i)" + maxMultKeyword);
				if (invSplit.length != 1){
					return new ProcessedAnnotationContext("",false,"Badly defined " + maxMultKeyword + "  keyword",null);
				}
				else{
					String contextName = invSplit[0].trim();
					return new ProcessedAnnotationContext(contextName,true,"Found max mult.",RestrictionType.MAXMULT);
				}
			}
			else {
				return new ProcessedAnnotationContext("",false,"Missing " + contextKeyword + "and" + maxMultKeyword + "  keywords",null);
			}
		}
	}
	
	
	private ProcessedAnnotationDetails getNoteDetails(Annotation note) {
		ProcessedAnnotationDetails result;
		String noteText = note.getText();
		String [] splitContextContent = noteText.split(contextContentSep);
		if (splitContextContent.length != 2){
			result = new ProcessedAnnotationDetails(false,"Badly defined expression. Revise the use of " + contextContentSep);
		}
		else{
			EventDepCondType evDepCond = null;
			String details = splitContextContent[1].replaceAll("\\s",""); //We eliminate any whitespaces, tabs, etc. to avoid issues
			if (details.contains(evDepCondImp)){
				String [] evDepSplit = details.split(evDepCondImp);
				if (evDepSplit.length != 2){
					result = new ProcessedAnnotationDetails(false,"Badly defined expression. There should only be one ->");
				}
				else {
					if(evDepSplit[0].equals("success") || evDepSplit[0].equals("succeed")){
						evDepCond = EventDepCondType.success;
					}
					else if(evDepSplit[0].equals("failure") || evDepSplit[0].equals("fail")){
						evDepCond = EventDepCondType.fail;
					}
					else{
						result = new ProcessedAnnotationDetails(false,"Condition should contain success or fail");
						return result;
					}
				}
			}
			if (details.toLowerCase().contains(selfRefKeyword.toLowerCase())){
				String [] selfSplit = details.split("(?i)" + selfRefKeyword);
				if(selfSplit.length != 2){
					result = new ProcessedAnnotationDetails(false,"Badly defined expression. Revise the use of " + selfRefKeyword);
					return result;
				}
				else{
					details = selfSplit[1];
				}
			}
			String sepSplit [] = details.split("\\" + elSeparator);
			if(sepSplit.length != 3){
				result = new ProcessedAnnotationDetails(false,"Badly defined expression. Revise the use of " + elSeparator
						+". Plus, we only allow expresions of the form 'self.targetClass.size' or 'targetClass.size'");
			}
			else {
				if(sepSplit[2].toLowerCase().contains(sizeKeyword.toLowerCase())){
					String elName = sepSplit[1];
					//Tractar el operands i garantir que size o size() apareixen sense mes text!
					String regex = "(?i)" + sizeKeyword + "|" + sizeKeyword + "\\(\\)";
					String sizeSplit [] = sepSplit[2].split("(?i)" + sizeKeyword + "\\(\\)" + "|" + sizeKeyword); //This can break VERY easily
					System.out.println("Size split length " + sizeSplit.length);
					if(sizeSplit.length != 2){
						result = new ProcessedAnnotationDetails(false, "Badly defined expression");
					}
					else{
						String splitChar;
						OpType opType;
						if (sizeSplit[1].contains("+")){
							splitChar = "\\+";
							opType = OpType.INCREMENTS;
						}
						else if (sizeSplit[1].contains("-")){
							splitChar = "\\-";
							opType = OpType.DECREMENTS;
						}
						else if (sizeSplit[1].contains("<")){
							if (sizeSplit[1].contains("=")){
								splitChar = "<=";
								opType = OpType.LESS_EQUALS;
							}
							else{
								splitChar = "<";
								opType = OpType.LESS_THAN;
							}
						}
						else if (sizeSplit[1].contains(">")){
							if (sizeSplit[1].contains("=")){
								splitChar = ">=";
								opType = OpType.GREATER_EQUALS;
							}
							else{
								splitChar = ">";
								opType = OpType.GREATER_THAN;
							}
						}
						else if (sizeSplit[1].contains("=")){
							splitChar = "=";
							opType = OpType.EQUALS;
						}
						else{
							opType = null;
							result = new ProcessedAnnotationDetails(false, "Expression should contain +,-,=,<=,>= ");
							return result; //Error
						}
						String [] symbolSplit = sizeSplit[1].split(splitChar);
						if (symbolSplit.length != 2){
							result = new ProcessedAnnotationDetails(false, "Expression contains more than one arithmetic operation (+,-,=,<=,>= )");
						}
						else {
							String numValue = symbolSplit[1];
							try {
								int num = Integer.parseInt(numValue);
								result = new ProcessedAnnotationDetails(elName, true, "Everything OK", opType, num, evDepCond);
							}
							catch (NumberFormatException numberExc){
								result = new ProcessedAnnotationDetails(false, "Expression " + numValue + " is not a valid number");
							}
							
						}
					}
				}
				else {
					result = new ProcessedAnnotationDetails(false, "Expression should contain " + sizeKeyword);
				}
			}
		}
		return result;
	}
	
	
	private class ProcessedAnnotationContext{
		private String m_context;
		private boolean m_successful;
		private String m_message;
		private RestrictionType m_type;
	
		private ProcessedAnnotationContext(String context, boolean successful, String message, RestrictionType type){
			m_context = context;
			m_successful = successful;
			m_message = message;
			m_type = type;
		}
		
		private boolean isSuccessful(){
			return m_successful;
		}
		
		private String getContext(){
			return m_context;
		}
		
		private RestrictionType getRestrictionType(){
			return m_type;
		}
		
	}
	
	private class ProcessedAnnotationDetails{
		private String m_targetArtifact;
		private boolean m_successful;
		private String m_message;
		private OpType m_opType;
		private int m_value;
		private EventDepCondType m_evDepCond; //If it is associated to an event-dependent condition.
		
		private ProcessedAnnotationDetails(String targetClass, boolean successful, String message, OpType opType, int value){
			m_targetArtifact = targetClass;
			m_successful = successful;
			m_message = message;
			m_opType = opType;
			m_value = value;
			m_evDepCond = null; //default value
		}
		
		private ProcessedAnnotationDetails(String targetClass, boolean successful, String message, OpType opType, int value, EventDepCondType evDepCond){
			m_targetArtifact = targetClass;
			m_successful = successful;
			m_message = message;
			m_opType = opType;
			m_value = value;
			m_evDepCond = evDepCond; 
		}
		
		/**
		 * Normally used when the result has been unsuccessful
		 * @param successful
		 * @param message
		 */
		private ProcessedAnnotationDetails(boolean successful, String message){
			m_targetArtifact = "";
			m_successful = successful;
			m_message = message;
			m_opType = null;
		}
		
		
		private boolean isSuccessful(){
			return m_successful;
		}
		
		private String getMessage(){
			return m_message;
		}
		
		private String getTargetArtifact(){
			return m_targetArtifact;
		}
		
		private OpType getOpType(){
			return m_opType;
		}
		
		private EventDepCondType getEventDepCond(){
			return m_evDepCond;
		}
		
		private int getValue(){
			return m_value;
		}
		
	}


}
