package tests;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.LinkedList;
import java.util.List;

import org.processmininguc.balsa.models.ActivityDiagram;
import org.processmininguc.balsa.models.ActivityDiagramActivity;
import org.processmininguc.balsa.models.Artifact;
import org.processmininguc.balsa.models.Balsa;
import org.processmininguc.balsa.models.StateMachineDiagram;
import org.processmininguc.balsa.models.StateMachineDiagramState;
import org.processmininguc.balsa.models.StateMachineDiagramTransition;
import org.processmininguc.balsa.models.impl.smadu.ActivityDiagramFactory;
import org.processmininguc.balsa.models.impl.smadu.ActivityDiagramImpl;
import org.processmininguc.balsa.models.impl.smadu.AnnotationFactory;
import org.processmininguc.balsa.models.impl.smadu.BalsaFactorySMADU;
import org.processmininguc.balsa.models.impl.smadu.StateMachineDiagramArtCondImpl;
import org.processmininguc.balsa.models.impl.smadu.StateMachineDiagramFactory;
import org.processmininguc.balsa.models.impl.smadu.StateMachineDiagramImpl;
import org.processmininguc.balsa.models.impl.smadu.StateMachineDiagramRealStateImpl;
import org.processmininguc.balsa.models.structures.ArtifactMultRel;
import org.processmininguc.balsa.models.structures.SimpleADChange;
import org.processmininguc.parser.AnnotationParser;
import org.processmininguc.utils.Utils;

import activity.Activity;
import parser.ParserError;
import parser.ParserErrorType;
import parser.ParserFacade;
import stateMachine.StateMachine;

public class ParserTest {
	
	
	public static void main(String[] args) throws Exception {
		
		//TODO: Aixo es un test senzill i hardcoded
		
		//String inputFile = "V:\\Dropbox\\PHD\\37-Tool\\argoUML\\prova-2estereotips-profile.xmi";
		String inputFile = "V:\\Dropbox\\PHD\\44-Bicing\\XML-exports\\project.xml";
		//String outputPath = "V:\\Dropbox\\PHD\\37-Tool\\proves\\result3";
		File file = new File(inputFile);
		
		InputStream is = new FileInputStream("V:\\Dropbox\\PHD\\44-Bicing\\XML-exports\\project.xml");
		
		BalsaFactorySMADU balsaFactory = new BalsaFactorySMADU();
		StateMachineDiagramFactory smdFactory = new StateMachineDiagramFactory();
		ActivityDiagramFactory adFactory = new ActivityDiagramFactory();
		AnnotationFactory annotFactory = new AnnotationFactory();
		
		List<StateMachine> smdList = new LinkedList<StateMachine>();
		List<Activity> adList = new LinkedList<Activity>();
		
		File tempFile = File.createTempFile("tempFile",".tmp");
        OutputStream outStream = new FileOutputStream(tempFile);
        byte[] buffer = new byte[is.available()];
        is.read(buffer);
        outStream.write(buffer);
        outStream.flush();
        outStream.close();
        is.close();
        
		ParserError peSMD = ParserFacade.parseStateMachines(tempFile, smdList, smdFactory);
		if (!peSMD.getErrorType().equals(ParserErrorType.SUCCESS)){
			//TODO: Improve this
			System.err.println("Error SMDs");
		}
		ParserError peAD = ParserFacade.parseActivityDiagrams(tempFile, adList, adFactory, annotFactory);
		if(!peAD.getErrorType().equals(ParserErrorType.SUCCESS)){
			//TODO: Improve this
			System.err.println("Error ADs");
		}
 		
		List<ActivityDiagramImpl> adListSMADU = Utils.convert(adList);
		List<StateMachineDiagramImpl> smdListSMADU = Utils.convert(smdList);
		
		List<StateMachineDiagram> smdListI = Utils.convert(smdListSMADU);
		List<ActivityDiagram> adListI = Utils.convert(adListSMADU);
		
			
		for(StateMachineDiagram smd: smdListI){
			//StateMachineDiagramImpl smdImp = (StateMachineDiagramImpl)smd;
			System.out.println("State Machine Diagram Name: " + smd.getName());
			List<StateMachineDiagramState> stateList = smd.getStateList();
			System.out.println("-------------");
			for (StateMachineDiagramState state: stateList){
				//StateMachineDiagramStateInt stateInt = (StateMachineDiagramStateInt) state;
				//System.out.println("State " + stateImp.get);
				List<StateMachineDiagramState> stateList2 = state.getNextStates();
				if (state.isState()){
					//State st = (State) stateInt;
					System.out.println("State " + ((StateMachineDiagramRealStateImpl) state).getName());
				}
				else if (state.isInitialPseudostate()){
					System.out.println("Initial state");
				}
				else if (state.isFinalState()){
					System.out.println("Final state");
				}
				else {
					System.out.println("Cannot deal with this type of state");
				}
			}
			System.out.println("------------");
			List<StateMachineDiagramTransition> transList = smd.getTransitionList();
			for (StateMachineDiagramTransition trans: transList){
				if (trans.getGuardCondition() != null){
					if (trans.getGuardCondition().referencesArtifactNum()){
						System.out.println("References another artifact: ");
						StateMachineDiagramArtCondImpl artCondImpl = (StateMachineDiagramArtCondImpl) trans.getGuardCondition();
						
						System.out.println("       Target artifact: " + artCondImpl.getTargetArtifact());
						System.out.println("       Number: " + artCondImpl.getNumber());
						System.out.println("       math op: " + artCondImpl.getMathOp().toString());
						System.out.println("*************************************");
					}
					else{
						System.out.println("Condition for transition: " + trans.getGuardCondition().getCondition());
					}
				}
			}
		}
		
		for (ActivityDiagram ad: adListI){
			System.out.println("Activity Diagram: " + ad.getName());
			System.out.println("-------------");
			//ActivityDiagramImpl adImpl = (ActivityDiagramImpl) ad;
			List<ActivityDiagramActivity> actList = ad.getActivities();
			//Set<Action> actList = adImpl.getActions();
			for(ActivityDiagramActivity act: actList){
				System.out.println("Task: " + act.getName());
			}
			ad.getNotes();
			System.out.println("-------------");
		}
		
		Balsa balsa = balsaFactory.createBalsa(smdListI, adListI);
		
		List<Artifact> artList = balsa.getArtifacts();
		
		for(Artifact art: artList){
			System.out.println("Artifact Name: " + art.getName());
			List<ArtifactMultRel> artMultRelList = art.getContext();
			if (!artMultRelList.isEmpty()){
				System.out.println("   - Context - ");
			}
			for(ArtifactMultRel artMultRel: artMultRelList){
				System.out.println("            Artifact " + artMultRel.getArtifactName());
				System.out.println("            Max. Card. " + artMultRel.getMaxCard());
				System.out.println("**************************");
			}
		}
		
		//Processing/Testing notes
		for (ActivityDiagram ad: adListI){
			System.out.println("AD: " + ad.getName());
			List<org.processmininguc.balsa.models.Annotation> noteList = ad.getNotes();
//			for (org.processmininguc.balsa.models.Annotation n : noteList){
//				System.out.println("HOOLAA" + n.getName());
//			}
			List<SimpleADChange> changeList = ad.getSimpleChanges();
			for(SimpleADChange change: changeList){
				System.out.println("Change List:");
				System.out.println("Context: " + change.getContext());
				System.out.println("Artifact: " + change.getTargetArtifact());
				System.out.println("Type: " + change.getOperationType());
				System.out.println("Number: " + change.getNumber());
				System.out.println("Event-dep. cond. : " + change.getEventDepCondType());
				System.out.println("**************************");
			}
			try {
				for(org.processmininguc.balsa.models.Annotation n: noteList){
					AnnotationParser np = new AnnotationParser();
					np.parseAnnotation(n);
					System.out.println("Note - Context: ");
				}
			}
			catch (Exception e){
				e.printStackTrace();
			}
			System.out.println("--------------------------------");

		}
		
		//System.out.println("Hola");
		
	
	}

}
